__author__ = "Jeremy Lainé"
__email__ = "jeremy.laine@m4x.org"
__license__ = "BSD"
__summary__ = "An implementation of WebRTC and ORTC"
__title__ = "aiortc"
__uri__ = "https://github.com/aiortc/aiortc"
__version__ = "1.0.0"
